// Cartesian tree - fast sliding window minimum search
// --------------------------------------------------------------------------------

#include <iostream>
#include <cstdlib>

class CTree {
private:
    class CTreeNode {
    public:
        CTreeNode *left, *right;
        CTreeNode *parent;
        CTreeNode *prev, *next;
        int value;
	CTreeNode () {
	    left = right = NULL;
	    parent = NULL;
	    prev = next = NULL;
	}
    };
    CTreeNode *root;
    CTreeNode *first, *last;
    int size;    
    int ncalls;
    int nsteps;
public:
    void addLeft(int value);
    void addRight(int value);
    void remLeft(void);
    void remRight(void);
    int getMin(void);
    int getSize(void);
    double getAverageStepsPerCall() {
        return ((double)nsteps)/ncalls; 
    }
    void print(CTreeNode *node, int level);
    void print(void);
    void clear(CTreeNode *node);
    CTree() {
	root = first = last = NULL;
	size = 0;
	ncalls = 0;
	nsteps = 0;
    }
    ~CTree() {
	clear(root);
    }
};

void CTree::addLeft(int value) {
    CTreeNode *newn = new CTreeNode();
    newn->value = value;
    ncalls++;
    nsteps++;
    if (size == 0) {
        root = first = last = newn;
    } else if (first->value <= value) {
        first->left = newn;
        newn->parent = first;
        newn->next = first;
    } else {
        CTreeNode *node = first;
        while (node != NULL && node->value > value) {
            node = node->parent;
	    nsteps++;
        }
        if (node == NULL) {
            newn->right = root;
            root->parent = newn;
            root = newn;
            newn->next = first;
        } else {
            newn->right = node->left;
            node->left->parent = newn;
            node->left = newn;
            newn->parent = node;
            newn->next = first;
        }
    }
    if (size != 0) first->prev = newn;
    first = newn;
    size++;
}

void CTree::addRight(int value) {
    CTreeNode *newn = new CTreeNode();
    newn->value = value;
    ncalls++;
    nsteps++;
    if (size == 0) {
        root = first = last = newn;
    } else if (last->value <= value) {
        last->right = newn;
        newn->parent = last;
        newn->prev = last;
    } else {
        CTreeNode *node = last;
        while (node != NULL && node->value > value) {
            node = node->parent;
	    nsteps++;
        }
        if (node == NULL) {
            newn->left = root;
            root->parent = newn;
            root = newn;
            newn->prev = last;
        } else {
            newn->left = node->right;
            node->right->parent = newn;
            node->right = newn;
            newn->parent = node;
            newn->prev = last;
        }
    }
    if (size != 0) last->next = newn;
    last = newn;
    size++;
}

void CTree::remLeft(void) {
    if (size == 0) return;
    if (size == 1) {
        delete root;
        root = first = last = NULL;
    } else if (first == root) {
        CTreeNode *node;
        root = first->right;
        root->parent = NULL;
        node = first->next;
        first->next->prev = NULL;
        delete first;
        first = node;      
    } else if (first->right == NULL) {
        CTreeNode *node;
        first->parent->left = NULL;
        node = first->next;
        first->next->prev = NULL;
        delete first;
        first = node;
    } else {
        CTreeNode *node;
        first->parent->left = first->right;
        first->right->parent = first->parent;
        node = first->next;
        first->next->prev = NULL;
        delete first;
        first = node;
    }
    size--;
}

void CTree::remRight(void) {
    if (size == 0) return;
    if (size == 1) {
        delete root;
        root = first = last = NULL;
    } else if (last == root) {
        CTreeNode *node;
        root = last->left;
        root->parent = NULL;
        node = last->prev;
        last->prev->next = NULL;
        delete last;
        last = node;      
    } else if (last->left == NULL) {
        CTreeNode *node;
        last->parent->right = NULL;
        node = last->prev;
        last->prev->next = NULL;
        delete last;
        last = node;
    } else {
        CTreeNode *node;
        last->parent->right = last->left;
        last->left->parent = last->parent;
        node = last->prev;
        last->prev->next = NULL;
        delete last;
        last = node;
    }
    size--;
}


int CTree::getMin(void) {
    return root->value;
}

int CTree::getSize(void) {
    return size;
}

void CTree::print(CTreeNode *node, int level) {
    for (int i = 0; i < level; i++) std::cout << " ";
    if (node == NULL) {
        std::cout << "o" << "\n";
    } else {
        std::cout << node->value << "\n";
        print(node->left, level + 1);
        print(node->right, level + 1);
    }
}

void CTree::print(void) {
    print(root, 0);
    std::cout << "\n";
}

void CTree::clear(CTreeNode *node) {
    if (node == NULL) return;
    clear(node->left);
    clear(node->right);
    delete node;
}

/*int taulu[1000000];

int main(void) {
    CTree *c = new CTree();
    int n = 0;

    for (int i = 0; i < 1000000; i++) {
        taulu[i] = rand() % 1000;
    }
    for (int i = 0; i < 20; i++) {
        c->addRight(taulu[i]);
    }
    int p = 19, s = 10000;
    while (n < 500000) {
	n++;
        int min;
        for (int i = 0; i < s; i++) {
            if (p > 100000) break;
	    p++;
            //std::cout << p << "\n";
            c->addRight(taulu[p]);
            c->remLeft();
        }
        min = taulu[p];
        for (int i = p; i > p - 20; i--) {
            if (taulu[i] < min) min = taulu[i];
        }
        if (min != c->getMin()) std::cout << "!\n";
        s = rand() % 100;
        for (int i = 0; i < s; i++) {
            if (p < 100) break;
            p--;
            //std::cout << p << "\n";
            c->addLeft(taulu[p - 19]);
            c->remRight();
        }
        min = taulu[p];
        for (int i = p; i > p - 20; i--) {
            if (taulu[i] < min) min = taulu[i];
        }
        if (min != c->getMin()) std::cout << "!\n";
        s = rand() % 100;
    }
    delete c;

    return 0;
}*/
